 /*
  * UAE - The Un*x Amiga Emulator
  * 
  * [n]curses output. May be usefull for people with very poor configuration.
  * [ Hardly think so (800x600 is faster on my machine), but this is good fun.
  *   You should get locked up for this, though. FIXME: this file wants color
  *   support added to it. -- Bernd ]
  * 
  * Copyright 1997, Samuel Devulder.
  */

/****************************************************************************/

#include "sysconfig.h"
#include "sysdeps.h"

#include <ctype.h>
#include <signal.h>

/****************************************************************************/

#include "uae.h"
#include "config.h"
#include "options.h"
#include "memory.h"
#include "custom.h"
#include "readcpu.h"
#include "newcpu.h"
#include "xwin.h"
#include "keyboard.h"
#include "keybuf.h"
#include "disk.h"
#include "debug.h"
#include "gui.h"

#ifdef HAVE_NCURSES_H
#include <ncurses.h>
#else
#include <curses.h>
#endif

/****************************************************************************/

#define MAXGRAYCHAR 128

static chtype graychar[MAXGRAYCHAR];
static int maxc,max_graychar;
static int curses_on;

xcolnr xcolors[4096];

 /* Keyboard and mouse */

static int keystate[256];
static int keydelay = 20;

int buttonstate[3];
int lastmx, lastmy;
int newmousecounters;

struct vidbuf_description gfxvidinfo;

/****************************************************************************/

static void curses_init(void);
static void curses_exit(void);
static void init_colors(void);
static void curses_insert_disk(int);

/****************************************************************************/

static RETSIGTYPE sigbrkhandler(int foo)
{
    curses_exit();
    activate_debugger();
}

void setup_brkhandler(void)
{
    struct sigaction sa;
    sa.sa_handler = sigbrkhandler;
    sa.sa_flags = 0;
    sa.sa_flags = SA_RESTART;
    sigemptyset(&sa.sa_mask);
    sigaction(SIGINT, &sa, NULL);
}

/****************************************************************************/

static void init_graychar(void)
{
    chtype *p = graychar;
    chtype attrs;
    int i,j;
    char *fmt;

    attrs = termattrs();
    if(attrs & A_REVERSE) fmt = " .:=obIXMB^v^#MXIbo=:. ";
    else                  fmt = " .`'^^\",:;i!1Il+=tfjxznuvyZYXHUOQ0MWB";
    /*                          ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
    /* One could find a better pattern..                                */
    attrs = A_NORMAL;
    while(*fmt) {
        if(*fmt == '^') {
            ++fmt;
            switch(*fmt) {
                case 'v': case 'V': attrs ^= A_REVERSE; break;
                case 'b': case 'B': attrs ^= A_BOLD; break;
                case 'd': case 'D': attrs ^= A_DIM; break;
                case 'u': case 'U': attrs ^= A_UNDERLINE; break;
                case 'p': case 'P': attrs  = A_NORMAL; break;
                case '#': if(ACS_CKBOARD == ':')
                               *p++ = (attrs | '#');
                          else *p++ = (attrs | ACS_CKBOARD); break;
                default:  *p++ = (attrs | *fmt); break;
            }
            ++fmt;
        } else *p++ = (attrs | *fmt++);
        if(p >= graychar + MAXGRAYCHAR) break;
    }
    max_graychar = (p - graychar) - 1;
#if 0
    for(j=0;j<LINES;++j) {
        move(j,0);
        for(i=0;i<COLS;++i) addch(graychar[i % (max_graychar+1)]);
    }
    refresh();
    sleep(3);
#endif
}

/****************************************************************************/

static int getcol(int x, int y)
{
    int xs, xe, ys, ye, c, cm;
    cm = c = 0;
    xs = ((x+0)*gfx_requested_width)/COLS;
    xe = ((x+1)*gfx_requested_width)/COLS;
    ys = ((y+0)*gfx_requested_height)/LINES;
    ye = ((y+1)*gfx_requested_height)/LINES;
    for(y = ys; y < ye; ++y)
    for(x = xs; x < xe; ++x)
        {
            c  += (UBYTE)gfxvidinfo.bufmem[x + y*gfx_requested_width];
            ++cm;
        }
    return cm?c/cm:0;
}

/****************************************************************************/

static void flush_line_txt(int y)
{
    int x;
    move(y,0);
    for(x=0;x<COLS;++x)
        {
            int c;
            c = getcol(x,y);
            if(!no_xhair) c = maxc - c;
            addch(graychar[(c*max_graychar)/maxc]);
        }
}

/****************************************************************************/

__inline__ void flush_line(int y)
{
    if(y<0 || y>=gfx_requested_height) {
/*       printf("flush_line out of window: %d\n", y); */
       return;
    }
    if(!curses_on) return;
    flush_line_txt((y*LINES)/gfx_requested_height);
}

/****************************************************************************/

void flush_block (int ystart, int ystop)
{
    int y;
    if(!curses_on) return;
    ystart = (ystart * LINES)/gfx_requested_height;
    ystop  = (ystop  * LINES)/gfx_requested_height;
    for(y=ystart; y<=ystop; ++y) flush_line_txt(y);
}

/****************************************************************************/

void flush_screen (int ystart, int ystop)
{
    if(!debugging && !curses_on) {
        curses_init();
        flush_block(0, gfx_requested_height - 1);
    }
    refresh();
}

/****************************************************************************/

static void init_colors(void)
{
    int i;
    for(i=0;i<4096;++i) {
        int r,g,b;
        r =  i>>8;
        g = (i>>4)&15;
        b =  i&15;
        xcolors[i] = (77 * r + 151 * g + 28 * b)/16;
        if(xcolors[i] > maxc) maxc = xcolors[i];
    }
}

/****************************************************************************/

static void curses_init()
{
    initscr();cbreak();noecho();
    nonl();intrflush(stdscr, FALSE); keypad(stdscr, TRUE);
    nodelay(stdscr, TRUE);
    leaveok(stdscr, TRUE);

#ifdef NCURSES_MOUSE_VERSION
    mousemask(BUTTON1_PRESSED | BUTTON1_RELEASED |
              BUTTON2_PRESSED | BUTTON2_RELEASED |
              BUTTON3_PRESSED | BUTTON3_RELEASED |
              REPORT_MOUSE_POSITION, NULL);
#endif

    init_graychar();
    curses_on = 1;
}

/****************************************************************************/

static void curses_exit()
{
#ifdef NCURSES_MOUSE_VERSION
    mousemask(0, NULL);
#endif

    nocbreak(); echo(); nl(); intrflush(stdscr, TRUE); 
    keypad(stdscr, FALSE); nodelay(stdscr, FALSE); leaveok(stdscr, FALSE);
    endwin();
    curses_on = 0;
}

/****************************************************************************/

int graphics_init(void)
{
    int i;

    if(!debugging) {
        curses_init();
        fprintf(stderr,"Using %s.\n",longname());
    }

    /* we have a 320x256x8 pseudo screen */

    gfx_requested_width      = 320;
    gfx_requested_height     = 256;
    gfx_requested_lores      = 1;

    gfxvidinfo.maxlinetoscr  = gfx_requested_width;
    gfxvidinfo.maxline       = gfx_requested_height;
    gfxvidinfo.maxblocklines = 100; /* whatever */
    gfxvidinfo.pixbytes      = 1;
    gfxvidinfo.rowbytes      = gfxvidinfo.pixbytes * gfx_requested_width;
    gfxvidinfo.bufmem        = (char *)calloc(gfxvidinfo.rowbytes, gfx_requested_height+1);

    if(!gfxvidinfo.bufmem) {
        fprintf(stderr,"Not enough memory.\n");
        return 0;
    } 

    init_colors();

    buttonstate[0] = buttonstate[1] = buttonstate[2] = 0;
    for(i=0; i<256; i++)
        keystate[i] = 0;
    
    lastmx = lastmy = 0; 
    newmousecounters = 0;

    return 1;
}

/****************************************************************************/

void graphics_leave(void)
{
    curses_exit();
}

/****************************************************************************/

static int keycode2amiga(int ch)
{
    switch(ch) {
        case KEY_A1:    return AK_NP7;
        case KEY_UP:    return AK_NP8;
        case KEY_A3:    return AK_NP9;
        case KEY_LEFT:  return AK_NP4;
        case KEY_B2:    return AK_NP5;
        case KEY_RIGHT: return AK_NP6;
        case KEY_C1:    return AK_NP1;
        case KEY_DOWN:  return AK_NP2;
        case KEY_C3:    return AK_NP3;
        case KEY_ENTER: return AK_ENT;
        case 13:        return AK_RET;
        case ' ':       return AK_SPC;
        case 27:        return AK_ESC;
        default: return -1;
    }
}

/***************************************************************************/

void handle_events(void)
{
    int ch;
    int kc;

    /* Hack to simulate key release */
    for(kc = 0; kc < 256; ++kc) {
        if(keystate[kc]) if(!--keystate[kc]) record_key((kc << 1) | 1);
    }
    if(buttonstate[0]) --buttonstate[0];
    if(buttonstate[1]) --buttonstate[1];
    if(buttonstate[2]) --buttonstate[2];

    newmousecounters = 0;
    if(!curses_on) return;

    while((ch = getch())!=ERR) {
        if(ch == 12) {clearok(stdscr,TRUE);refresh();}
#ifdef NCURSES_MOUSE_VERSION
        if(ch == KEY_MOUSE) {
            MEVENT ev;
            if(getmouse(&ev) == OK) {
                lastmx = (ev.x*gfx_requested_width)/COLS;
                lastmy = (ev.y*gfx_requested_height)/LINES;
                if(ev.bstate & BUTTON1_PRESSED)  buttonstate[0] = keydelay;
                if(ev.bstate & BUTTON1_RELEASED) buttonstate[0] = 0;
                if(ev.bstate & BUTTON2_PRESSED)  buttonstate[1] = keydelay;
                if(ev.bstate & BUTTON2_RELEASED) buttonstate[1] = 0;
                if(ev.bstate & BUTTON3_PRESSED)  buttonstate[2] = keydelay;
                if(ev.bstate & BUTTON3_RELEASED) buttonstate[2] = 0;
            }
        }
#endif
        if(ch == 6)  ++lastmx; /* ^F */
        if(ch == 2)  --lastmx; /* ^B */
        if(ch == 14) ++lastmy; /* ^N */
        if(ch == 16) --lastmy; /* ^P */
	if(ch == 11) {buttonstate[0] = keydelay;ch = 0;} /* ^K */
	if(ch == 25) {buttonstate[2] = keydelay;ch = 0;} /* ^Y */
	
        if(ch >= KEY_F(1) && ch<=KEY_F(4)) {
	  curses_insert_disk(ch - KEY_F(1));
	  ch = 0;
	}

        if(isupper(ch)) {
            keystate[AK_LSH] =
            keystate[AK_RSH] = keydelay;
            record_key(AK_LSH << 1);
            record_key(AK_RSH << 1);
            kc = keycode2amiga(tolower(ch));
            keystate[kc] = keydelay;
            record_key(kc << 1);
        } else if((kc = keycode2amiga(ch)) >= 0) {
            keystate[kc] = keydelay;
            record_key(kc << 1);
        }
    }
    gui_handle_events();
}

/***************************************************************************/

void target_specific_usage(void)
{
    printf("----------------------------------------------------------------------------\n");
    printf("[n]curses specific usage:\n");
    printf("  -x : Display reverse video.\n");
    printf("By default uae will assume a black on white display. If yours\n");
    printf("is light on dark, use -x. In case of graphics garbage, ^L will\n");
    printf("redisplay the screen. ^K simulate left mouse button, ^Y RMB.\n");
    printf("If you are using a xterm UAE can use the mouse. Else use ^F ^B\n");
    printf("^P ^N to emulate mouse mouvements.\n");
    printf("----------------------------------------------------------------------------\n");
}

/***************************************************************************/

int debuggable(void)
{
    return 1;
}

/***************************************************************************/

int needmousehack(void)
{
    return 1;
}

/***************************************************************************/

void LED(int on)
{
}

/***************************************************************************/

static void curses_insert_disk(int dsk)
{
    char buff[256];
    char *s = buff;
    curses_exit();
    fprintf(stderr, "Old DF%d filename: %s\n", dsk, dsk==0?df0:dsk==1?df1:
                                                    dsk==2?df2:df3);
    fprintf(stderr,"Please type filename for drive %d: ", dsk);
    fflush(stderr);
    fgets(buff, 255, stdin);
    while(*s) if(*s == '\n') *s='\0'; else ++s;
    if(*buff) disk_insert(dsk, buff);
    flush_screen(0,0);
}

/***************************************************************************/


